@@ -65,11 +65,27 @@ static int issetugid() {
 }
 #endif
 
+#if defined(HAVE_IPV6) && defined(HAVE_INET_PTON)
+# define USE_IPV6
+#endif
+
+#ifdef USE_IPV6
+#define PACKAGE_FEATURES " (ipv6)"
+#else
+#define PACKAGE_FEATURES ""
+#endif
+
+#define PACKAGE_DESC "spawn-fcgi v" PACKAGE_VERSION PACKAGE_FEATURES " - spawns FastCGI processes\n"
+
+
 static int bind_socket(const char *addr, unsigned short port, const char *unixsocket, uid_t uid, gid_t gid, int mode) {
 	int fcgi_fd, socket_type, val;
 
 	struct sockaddr_un fcgi_addr_un;
 	struct sockaddr_in fcgi_addr_in;
+#ifdef USE_IPV6
+	struct sockaddr_in6 fcgi_addr_in6;
+#endif
 	struct sockaddr *fcgi_addr;
 
 	socklen_t servlen;
@@ -118,16 +134,40 @@ static int bind_socket(const char *addr, unsigned short port, const char *unixso
 	} else {
 		memset(&fcgi_addr_in, 0, sizeof(fcgi_addr_in));
 		fcgi_addr_in.sin_family = AF_INET;
-		if (addr != NULL) {
-			fcgi_addr_in.sin_addr.s_addr = inet_addr(addr);
-		} else {
-			fcgi_addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
-		}
 		fcgi_addr_in.sin_port = htons(port);
-		servlen = sizeof(fcgi_addr_in);
 
+		servlen = sizeof(fcgi_addr_in);
 		socket_type = AF_INET;
 		fcgi_addr = (struct sockaddr *) &fcgi_addr_in;
+
+#ifdef USE_IPV6
+		memset(&fcgi_addr_in6, 0, sizeof(fcgi_addr_in6));
+		fcgi_addr_in6.sin6_family = AF_INET6;
+		fcgi_addr_in6.sin6_port = fcgi_addr_in.sin_port;
+#endif
+
+		if (addr == NULL) {
+			fcgi_addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
+#ifdef HAVE_INET_PTON
+		} else if (1 == inet_pton(AF_INET, addr, &fcgi_addr_in.sin_addr)) {
+			/* nothing to do */
+#ifdef HAVE_IPV6
+		} else if (1 == inet_pton(AF_INET6, addr, &fcgi_addr_in6.sin6_addr)) {
+			servlen = sizeof(fcgi_addr_in6);
+			socket_type = AF_INET6;
+			fcgi_addr = (struct sockaddr *) &fcgi_addr_in6;
+#endif
+		} else {
+			fprintf(stderr, "spawn-fcgi: '%s' is not a valid IP address\n", addr);
+			return -1;
+#else
+		} else {
+			if ((in_addr_t)(-1) == (fcgi_addr_in.sin_addr.s_addr = inet_addr(addr))) {
+				fprintf(stderr, "spawn-fcgi: '%s' is not a valid IPv4 address\n", addr);
+				return -1;
+			}
+#endif
+		}
 	}
 
 
@@ -308,13 +348,14 @@ static int find_user_group(const char *user, const char *group, uid_t *uid, gid_
 	gid_t my_gid = 0;
 	struct passwd *my_pwd = NULL;
 	struct group *my_grp = NULL;
+	char *endptr = NULL;
 	*uid = 0; *gid = 0;
 	if (username) *username = NULL;
 
 	if (user) {
-		my_uid = strtol(user, NULL, 10);
+		my_uid = strtol(user, &endptr, 10);
 
-		if (my_uid <= 0) {
+		if (my_uid <= 0 || *endptr) {
 			if (NULL == (my_pwd = getpwnam(user))) {
 				fprintf(stderr, "spawn-fcgi: can't find user name %s\n", user);
 				return -1;
@@ -329,14 +370,14 @@ static int find_user_group(const char *user, const char *group, uid_t *uid, gid_
 			if (username) *username = user;
 		} else {
 			my_pwd = getpwuid(my_uid);
-			if (username) *username = my_pwd->pw_name;
+			if (username && my_pwd) *username = my_pwd->pw_name;
 		}
 	}
 
 	if (group) {
-		my_gid = strtol(group, NULL, 10);
+		my_gid = strtol(group, &endptr, 10);
 
-		if (my_gid <= 0) {
+		if (my_gid <= 0 || *endptr) {
 			if (NULL == (my_grp = getgrnam(group))) {
 				fprintf(stderr, "spawn-fcgi: can't find group name %s\n", group);
 				return -1;
@@ -362,7 +403,6 @@ static int find_user_group(const char *user, const char *group, uid_t *uid, gid_
 	return 0;
 }
 
-#define PACKAGE_DESC "spawn-fcgi v" PACKAGE_VERSION " - spawns FastCGI processes\n"
 static void show_version () {
 	const char *b = PACKAGE_DESC
 "Build-Date: " __DATE__ " " __TIME__ "\n";
@@ -374,12 +414,12 @@ static void show_help () {
 	char *b = \
 "Usage: spawn-fcgi [options] [-- <fcgiapp> [fcgi app arguments]]\n" \
 "\n" \
-"spawn-fcgi v" PACKAGE_VERSION " - spawns FastCGI processes\n" \
+PACKAGE_DESC \
 "\n" \
 "Options:\n" \
 " -f <path>      filename of the fcgi-application (ignored if <fcgiapp> is given)\n" \
 " -d <directory> chdir to directory before spawning\n" \
-" -a <address>   bind to IP address\n" \
+" -a <address>   bind to IPv4/IPv6 address (defaults to 0.0.0.0)\n" \
 " -p <port>      bind to TCP-port\n" \
 " -s <path>      bind to Unix domain socket\n" \
 " -M <mode>      change Unix domain socket mode\n" \
@@ -408,6 +448,7 @@ int main(int argc, char **argv) {
 	     *sockusername = NULL, *sockgroupname = NULL, *fcgi_dir = NULL,
 	     *addr = NULL;
 	char **fcgi_app_argv = { NULL };
+	char *endptr = NULL;
 	unsigned short port = 0;
 	int sockmode = -1;
 	int child_count = -1;
@@ -431,7 +472,12 @@ int main(int argc, char **argv) {
 		case 'f': fcgi_app = optarg; break;
 		case 'd': fcgi_dir = optarg; break;
 		case 'a': addr = optarg;/* ip addr */ break;
-		case 'p': port = strtol(optarg, NULL, 10);/* port */ break;
+		case 'p': port = strtol(optarg, &endptr, 10);/* port */
+			if (*endptr) {
+				fprintf(stderr, "spawn-fcgi: invalid port: %u\n", (unsigned int) port);
+				return -1;
+			}
+			break;
 		case 'C': child_count = strtol(optarg, NULL, 10);/*  */ break;
 		case 'F': fork_count = strtol(optarg, NULL, 10);/*  */ break;
 		case 's': unixsocket = optarg; /* unix-domain socket */ break;
@@ -526,6 +572,10 @@ int main(int argc, char **argv) {
 		if (-1 == find_user_group(sockusername, sockgroupname, &sockuid, &sockgid, NULL))
 			return -1;
 
+		if (uid != 0 && gid == 0) {
+			fprintf(stderr, "spawn-fcgi: WARNING: couldn't find the user for uid %i and no group was specified, so only the user privileges will be dropped\n", (int) uid);
+		}
+
 		if (0 == sockuid) sockuid = uid;
 		if (0 == sockgid) sockgid = gid;
 
